/*
    terminatorX - realtime audio scratching software
    Copyright (C) 1999, 2000 Alexander Knig
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
    File: tX_seqpar.cc
 
    Description: This implements the "sequenceable parameters".
*/    

#include "tX_seqpar.h"
#include "tX_vtt.h"
#include <stdio.h>
#include "tX_mastergui.h"
#include "tX_global.h"
#include "tX_sequencer.h"

list <tX_seqpar *> tX_seqpar :: all;
list <tX_seqpar *> tX_seqpar :: update;
pthread_mutex_t tX_seqpar :: update_lock = PTHREAD_MUTEX_INITIALIZER;

#define tt ((vtt_class *) vtt)

tX_seqpar :: tX_seqpar ()
{
	touched=0;
	gui_active=1;
	vtt=NULL;
	all.push_back(this);
	
	//printf("%i sequencable parameters registered.\n", all.size());
}

tX_seqpar :: tX_seqpar (void * mytt)
{
	vtt=mytt;
	touched=0;
	gui_active=1;
	
	all.push_back(this);
}

void tX_seqpar :: set_vtt (void *mytt)
{
	vtt=mytt;
}

tX_seqpar :: ~tX_seqpar()
{
	pthread_mutex_lock(&update_lock);
	update.remove(this);
	pthread_mutex_unlock(&update_lock);
	
	sequencer.delete_all_events_for_sp(this);
	all.remove(this);
}

void tX_seqpar :: do_touch()
{
	if (sequencer.is_recording())
	{
		touched=1;
		touch_timestamp=sequencer.get_timestamp();
	}
}

void tX_seqpar :: untouch_all()
{
	list <tX_seqpar *> :: iterator sp;
	
	for (sp=all.begin(); sp!=all.end(); sp++)
	{
		(*sp)->untouch();
	}
}

void tX_seqpar :: create_persistence_ids()
{
	list <tX_seqpar *> :: iterator sp;
	int pid=0;
	
	for (sp=all.begin(); sp!=all.end(); sp++)
	{
		pid++;
		(*sp)->set_persistence_id(pid);
	}
}

tX_seqpar* tX_seqpar :: get_sp_by_persistence_id(int pid)
{
	list <tX_seqpar *> :: iterator sp;
	
	for (sp=all.begin(); sp!=all.end(); sp++)
	{
		if ((*sp)->get_persistence_id()==pid) return ((*sp));
	}
	
	fprintf (stderr, "oops: failed to lookup persistence id [%i].\n", pid);	
	return (NULL);
}


void tX_seqpar :: record_value(const float value)
{
#define last_event ((tX_event *) last_event_recorded)

	/* recording more than one event per seqpar for
	  one timestamp doesn't make sense... so if the 
	  last_event_recorded was for the current timestamp
	  we simply set that event's value to the current one.
	*/
	if ((last_event) && (last_event->get_timestamp() == sequencer.get_timestamp()))
	{
		last_event->set_value(value);
	}
	else last_event_recorded=(void *) sequencer.record(this, value);
}

void tX_seqpar :: receive_gui_value(const float value)
{
	if (gui_active)
	{
		touch();
		do_exec(value);
		record_value(value);
	}
}

void tX_seqpar :: receive_input_value(const float value)
{
	touch();
	exec_value(value);
	record_value(value);
}

void tX_seqpar :: receive_forward_value(const float value)
{
	fwd_value=value;
}

void tX_seqpar :: materialize_forward_values()
{
	list <tX_seqpar *> :: iterator sp;
	
	for (sp=all.begin(); sp!=all.end(); sp++)
	{
		(*sp)->exec_value((*sp)->fwd_value);
	}	
	gdk_flush();
}

char * tX_seqpar :: get_vtt_name()
{       
        if (vtt) return tt->name;
        else return "Master Track";
}

char * tX_seqpar :: get_name()
{
        return "This string means trouble!";
}

void tX_seqpar :: do_exec(const float value)
{
	fprintf(stderr, "oops: called do_exec() of tX_seqpar.\n");
}

void tX_seqpar :: exec_value(const float value)
{
	fprintf(stderr, "oops: called exec_value() of tX_seqpar.\n");
}

void tX_seqpar :: do_update_graphics()
{
	fprintf(stderr, "oops: called do_update_graphics() of tX_seqpar.\n");
}

void tX_seqpar :: update_graphics()
{
	gui_active=0;
	do_update_graphics();
	while (gtk_events_pending()) gtk_main_iteration();	/* gtk_flush */
	gui_active=1;
}

void tX_seqpar :: update_all_graphics()
{
	list <tX_seqpar *> :: iterator sp;


	pthread_mutex_lock(&update_lock);

	if (!update.size())
	{
		pthread_mutex_unlock(&update_lock);
		return;	
	}
	
	while (gtk_events_pending()) gtk_main_iteration();	
	for (sp=update.begin(); sp!=update.end(); sp++)
	{
		(*sp)->update_graphics();
	}
	update.erase(update.begin(), update.end());
	pthread_mutex_unlock(&update_lock);
}

void tX_seqpar_update :: exec_value(const float value)
{
	do_exec(value);
	pthread_mutex_lock(&update_lock);
	update.push_front(this);
	pthread_mutex_unlock(&update_lock);
}

void tX_seqpar_no_update :: exec_value(const float value)
{
	do_exec(value);
}

void tX_seqpar_no_update_active_forward :: receive_forward_value(const float value)
{
	fwd_value=value;
	do_exec(value);
}	

void tX_seqpar_update_active_forward :: receive_forward_value(const float value)
{
	fwd_value=value;
	do_exec(value);
}

/* real classes */

void tX_seqpar_master_volume :: do_exec(const float value)
{
	vtt_class :: set_master_volume(value);
}

void tX_seqpar_master_volume :: do_update_graphics ()
{
	gtk_adjustment_set_value(volume_adj, vtt_class::master_volume);
}

char *tX_seqpar_master_volume :: get_name()
{
        return "Master Volume";
}

void tX_seqpar_master_pitch :: do_exec(const float value)
{
	vtt_class :: set_master_pitch(value);
}

void tX_seqpar_master_pitch :: do_update_graphics ()
{
	gtk_adjustment_set_value(pitch_adj, globals.pitch);
}

char *tX_seqpar_master_pitch :: get_name()
{
        return "Master Pitch";
}

void tX_seqpar_vtt_speed :: do_exec(const float value)
{
	tt->speed=value;
}

char *tX_seqpar_vtt_speed :: get_name()
{
        return "Speed (Scratching)";
}

void tX_seqpar_spin :: do_exec(const float value)
{
	if (value > 0) tt->speed=tt->res_pitch;
	else tt->speed=0;
}

char *tX_seqpar_spin :: get_name()
{
        return "Motor Spin (On/Off)";
}

void tX_seqpar_vtt_volume :: do_exec(const float value)
{
	tt->set_volume(value);
}

void tX_seqpar_vtt_volume :: do_update_graphics ()
{
	gtk_adjustment_set_value(tt->gui.volume, 2.0-tt->rel_volume);
}

char *tX_seqpar_vtt_volume :: get_name()
{
        return "Volume";
}

void tX_seqpar_vtt_pitch :: do_exec(const float value)
{
	tt->set_pitch(value);
}

void tX_seqpar_vtt_pitch :: do_update_graphics ()
{
	gtk_adjustment_set_value(tt->gui.pitch, tt->rel_pitch);
}

char *tX_seqpar_vtt_pitch :: get_name()
{
        return "Pitch";
}

void tX_seqpar_vtt_trigger :: do_exec(const float value)
{
	if (value > 0) tt->trigger();
	else tt->stop();
}

char *tX_seqpar_vtt_trigger :: get_name()
{
        return "Trigger (Start/Stop)";
}

void tX_seqpar_vtt_loop :: do_exec(const float value)
{
	tt->set_loop(value>0);
}

char *tX_seqpar_vtt_loop :: get_name()
{
        return "Loop (On/Off)";
}

void tX_seqpar_vtt_loop :: do_update_graphics ()
{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tt->gui.loop), tt->loop);
}

void tX_seqpar_vtt_sync_client :: do_exec(const float value)
{
	tt->set_sync_client((value>0), tt->sync_cycles);
}

void tX_seqpar_vtt_sync_client :: do_update_graphics ()
{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tt->gui.sync_client), tt->is_sync_client);
}

char *tX_seqpar_vtt_sync_client :: get_name()
{
        return "Sync Client (On/Off)";
}

void tX_seqpar_vtt_sync_cycles :: do_exec(const float value)
{
	tt->set_sync_client(tt->is_sync_client, (int) value);
}

void tX_seqpar_vtt_sync_cycles :: do_update_graphics ()
{
	gtk_adjustment_set_value(tt->gui.cycles, tt->sync_cycles);
}

char *tX_seqpar_vtt_sync_cycles :: get_name()
{
        return "Sync Cycles";
}

void tX_seqpar_vtt_lp_enable :: do_exec(const float value)
{
	tt->lp_set_enable(value>0);
}

void tX_seqpar_vtt_lp_enable :: do_update_graphics ()
{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tt->gui.lp_enable), tt->lp_enable);
}

char *tX_seqpar_vtt_lp_enable :: get_name()
{
        return "Lowpass: Enable (On/Off)";
}

void tX_seqpar_vtt_lp_gain :: do_exec(const float value)
{
	tt->lp_set_gain(value);
}

char *tX_seqpar_vtt_lp_gain :: get_name()
{
        return "Lowpass: Input Gain";
}

void tX_seqpar_vtt_lp_gain :: do_update_graphics ()
{
	gtk_adjustment_set_value(tt->gui.lp_gain, tt->lp_gain);
}

void tX_seqpar_vtt_lp_reso :: do_exec(const float value)
{
	tt->lp_set_reso(value);
}

void tX_seqpar_vtt_lp_reso :: do_update_graphics ()
{
	gtk_adjustment_set_value(tt->gui.lp_reso, tt->lp_reso);
}

char *tX_seqpar_vtt_lp_reso :: get_name()
{
        return "Lowpass: Resonance";
}

void tX_seqpar_vtt_lp_freq :: do_exec(const float value)
{
	tt->lp_set_freq(value);
}

char *tX_seqpar_vtt_lp_freq :: get_name()
{
        return "Lowpass: Cutoff Frequency";
}

void tX_seqpar_vtt_lp_freq :: do_update_graphics ()
{
	gtk_adjustment_set_value(tt->gui.lp_freq, tt->lp_freq);
}

void tX_seqpar_vtt_ec_enable :: do_exec(const float value)
{
	tt->ec_set_enable(value>0);
}

void tX_seqpar_vtt_ec_enable :: do_update_graphics ()
{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tt->gui.ec_enable), tt->ec_enable);
}

char *tX_seqpar_vtt_ec_enable :: get_name()
{
        return "Echo: Enable (On/Off)";
}

void tX_seqpar_vtt_ec_length :: do_exec(const float value)
{
	tt->ec_set_length(value);
}

void tX_seqpar_vtt_ec_length :: do_update_graphics ()
{
	gtk_adjustment_set_value(tt->gui.ec_length, tt->ec_length);
}

char *tX_seqpar_vtt_ec_length :: get_name()
{
        return "Echo: Duration";
}

void tX_seqpar_vtt_ec_feedback :: do_exec(const float value)
{
	tt->ec_set_feedback(value);
}

void tX_seqpar_vtt_ec_feedback :: do_update_graphics ()
{
	gtk_adjustment_set_value(tt->gui.ec_feedback, tt->ec_feedback);
}

char *tX_seqpar_vtt_ec_feedback :: get_name()
{
        return "Echo: Feedback";
}

void tX_seqpar_vtt_mute :: do_exec(const float value)
{
	tt->set_mute(value>0);
}

char *tX_seqpar_vtt_mute :: get_name()
{
        return "Mute (On/Off)";
}

